#include <stdexcept>
#include <cmath>
#include <msp/gl/extensions/msp_legacy_features.h>
#include "error.h"
#include "light.h"
#include "lighting.h"
#include "lightunit.h"
#include "matrix.h"
#include "misc.h"

using namespace std;

namespace Msp {
namespace GL {

Lighting::Lighting():
	ambient(0.2),
	zenith_direction(0, 0, 1),
	horizon_angle(Geometry::Angle<float>::zero()),
	fog_color(0.0f, 0.0f, 0.0f, 0.0f),
	fog_density(0.0f)
{ }

Lighting::~Lighting()
{
	for(vector<Light *>::iterator i=owned_data.begin(); i!=owned_data.end(); ++i)
		delete *i;
}

void Lighting::set_ambient(const Color &a)
{
	ambient = a;
}

void Lighting::set_sky_color(const Color &s)
{
	sky_color = s;
}

void Lighting::set_zenith_direction(const Vector3 &d)
{
	zenith_direction = d;
}

void Lighting::set_horizon_angle(const Geometry::Angle<float> &a)
{
	horizon_angle = a;
}

void Lighting::set_fog_color(const Color &c)
{
	fog_color = c;
}

void Lighting::set_fog_density(float d)
{
	if(d<0)
		throw invalid_argument("Lighting::set_fog_density");

	fog_density = d;
}

void Lighting::set_fog_half_distance(float d)
{
	set_fog_density(-log(pow(0.5, 1.0/d)));
}

void Lighting::attach(unsigned i, const Light &l)
{
	if(i>=lights.size())
		lights.resize(i+1);

	lights[i] = &l;
	if(current()==this)
		l.bind_to(i);
}

void Lighting::detach(unsigned i)
{
	if(i>=lights.size())
		return;

	lights[i] = 0;
	if(current()==this)
		Light::unbind_from(i);
}

const Light *Lighting::get_attached_light(unsigned i) const
{
	return i<lights.size() ? lights[i] : 0;
}

void Lighting::update_shader_data(ProgramData &shdata, const Matrix &view_matrix) const
{
	shdata.uniform("ambient_color", ambient);
	shdata.uniform("sky_color", sky_color);
	shdata.uniform("eye_zenith_dir", view_matrix.block<3, 3>(0, 0)*zenith_direction);
	shdata.uniform("horizon_limit", horizon_angle.radians());
	shdata.uniform("fog_color", fog_color);
	shdata.uniform("fog_density", fog_density);

	// For backwards compatibility
	shdata.uniform("eye_sky_dir", view_matrix.block<3, 3>(0, 0)*zenith_direction);

	for(unsigned i=0; i<lights.size(); ++i)
		if(lights[i])
			lights[i]->update_shader_data(shdata, view_matrix, i);
}

void Lighting::bind() const
{
	static Require _req(MSP_legacy_features);
	if(lights.size()>LightUnit::get_n_units())
		throw invalid_operation("Lighting::bind");

	const Lighting *old = current();
	if(!set_current(this))
		return;

	enable(GL_LIGHTING);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, &ambient.r);
	for(unsigned i=0; i<lights.size(); ++i)
	{
		if(lights[i])
			lights[i]->bind_to(i);
		else
			Light::unbind_from(i);
	}

	if(old)
	{
		for(unsigned i=lights.size(); i<old->lights.size(); ++i)
			Light::unbind_from(i);
	}

	if(fog_density)
	{
		enable(GL_FOG);
		glFogi(GL_FOG_MODE, GL_EXP);
		glFogf(GL_FOG_DENSITY, fog_density);
		glFogfv(GL_FOG_COLOR, &fog_color.r);
	}
}

void Lighting::unbind()
{
	const Lighting *old = current();
	if(!set_current(0))
		return;

	for(unsigned i=0; i<old->lights.size(); ++i)
		if(old->lights[i])
			Light::unbind_from(i);

	disable(GL_LIGHTING);
	if(old->fog_density)
		disable(GL_FOG);
}


Lighting::Loader::Loader(Lighting &l):
	DataFile::ObjectLoader<Lighting>(l)
{
	add("ambient", &Loader::ambient);
	add("fog_color", &Loader::fog_color);
	add("fog_density", &Loader::fog_density);
	add("fog_half_distance", &Loader::fog_half_distance);
	add("horizon_angle", &Loader::horizon_angle);
	add("light", &Loader::light);
	add("sky_color", &Loader::sky_color);
	add("zenith_direction", &Loader::zenith_direction);
}

void Lighting::Loader::ambient(float r, float g, float b)
{
	obj.ambient = Color(r, g, b);
}

void Lighting::Loader::fog_color(float r, float g, float b)
{
	obj.set_fog_color(Color(r, g, b));
}

void Lighting::Loader::fog_density(float d)
{
	obj.set_fog_density(d);
}

void Lighting::Loader::fog_half_distance(float d)
{
	obj.set_fog_half_distance(d);
}

void Lighting::Loader::horizon_angle(float a)
{
	obj.set_horizon_angle(Geometry::Angle<float>::from_degrees(a));
}

void Lighting::Loader::light(unsigned i)
{
	RefPtr<Light> lgt = new Light;
	load_sub(*lgt);
	obj.attach(i, *lgt);
	obj.owned_data.push_back(lgt.release());
}

void Lighting::Loader::sky_color(float r, float g, float b)
{
	obj.set_sky_color(Color(r, g, b));
}

void Lighting::Loader::zenith_direction(float x, float y, float z)
{
	obj.set_sky_direction(Vector3(x, y, z));
}

} // namespace GL
} // namespace Msp
